##############################################################################
#
# Copyright (c) 2001, 2002 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Site error log module.
"""

import logging
import os
import sys
import time
from random import random

from AccessControl.class_init import InitializeClass
from AccessControl.SecurityInfo import ClassSecurityInfo
from AccessControl.SecurityManagement import getSecurityManager
from AccessControl.unauthorized import Unauthorized
from Acquisition import aq_acquire
from Acquisition import aq_base
from App.Dialogs import MessageDialog
from OFS.SimpleItem import SimpleItem
from Products.PageTemplates.PageTemplateFile import PageTemplateFile
from zExceptions.ExceptionFormatter import format_exception
from zope.component import adapter
from zope.event import notify
from ZPublisher.interfaces import IPubFailure

from .interfaces import ErrorRaisedEvent


try:
    # Python 3
    from _thread import allocate_lock
except ImportError:
    # Python 2
    from thread import allocate_lock


LOG = logging.getLogger('Zope.SiteErrorLog')

# Permission names
use_error_logging = 'Log Site Errors'
log_to_event_log = 'Log to the Event Log'

# We want to restrict the rate at which errors are sent to the Event Log
# because we know that these errors can be generated quick enough to
# flood some zLOG backends. zLOG is used to notify someone of a problem,
# not to record every instance.
# This dictionary maps exception name to a value which encodes when we
# can next send the error with that name into the event log. This dictionary
# is shared between threads and instances. Concurrent access will not
# do much harm.
_rate_restrict_pool = {}

# The number of seconds that must elapse on average between sending two
# exceptions of the same name into the the Event Log. one per minute.
_rate_restrict_period = 60

# The number of exceptions to allow in a burst before the above limit
# kicks in. We allow five exceptions, before limiting them to one per
# minute.
_rate_restrict_burst = 5

_www = os.path.join(os.path.dirname(__file__), 'www')

# temp_logs holds the logs.
temp_logs = {}  # { oid -> [ traceback string ] }

cleanup_lock = allocate_lock()

try:
    # Python 2
    bstr = basestring
except NameError:
    # Python 3
    bstr = str


class SiteErrorLog(SimpleItem):
    """Site error log class.  You can put an error log anywhere in the tree
    and exceptions in that area will be posted to the site error log.
    """
    meta_type = 'Site Error Log'
    id = 'error_log'
    zmi_icon = 'fas fa-bug'
    zmi_show_add_dialog = False

    keep_entries = 20
    copy_to_zlog = True

    security = ClassSecurityInfo()

    manage_options = (
        {'label': 'Log', 'action': 'manage_main'},
    ) + SimpleItem.manage_options

    security.declareProtected(use_error_logging, 'manage_main')  # NOQA: D001
    manage_main = PageTemplateFile('main.pt', _www)

    security.declareProtected(use_error_logging, 'showEntry')  # NOQA: D001
    showEntry = PageTemplateFile('showEntry.pt', _www)

    @security.private
    def manage_beforeDelete(self, item, container):
        if item is self:
            try:
                del container.__error_log__
            except AttributeError:
                pass

    @security.private
    def manage_afterAdd(self, item, container):
        if item is self:
            container.__error_log__ = aq_base(self)

    def _setId(self, id):
        if id != self.id:
            raise ValueError(MessageDialog(
                title='Invalid Id',
                message='Cannot change the id of a SiteErrorLog',
                action='./manage_main'))

    def _getLog(self):
        """Returns the log for this object.

        Careful, the log is shared between threads.
        """
        log = temp_logs.get(self._p_oid, None)
        if log is None:
            log = []
            temp_logs[self._p_oid] = log
        return log

    @security.protected(use_error_logging)
    def forgetEntry(self, id, REQUEST=None):
        """Removes an entry from the error log."""
        log = self._getLog()
        cleanup_lock.acquire()
        i = 0
        for entry in log:
            if entry['id'] == id:
                del log[i]
            i += 1
        cleanup_lock.release()
        if REQUEST is not None:
            REQUEST.RESPONSE.redirect(
                '%s/manage_main?manage_tabs_message='
                'Error+log+entry+was+removed.' %
                self.absolute_url())

    # Exceptions that happen all the time, so we dont need
    # to log them. Eventually this should be configured
    # through-the-web.
    _ignored_exceptions = ('Unauthorized', 'NotFound', 'Redirect')

    @security.private
    def raising(self, info):
        """Log an exception.

        Called by SimpleItem's exception handler.
        Returns the url to view the error log entry
        """
        now = time.time()
        try:
            tb_text = None
            tb_html = None

            strtype = str(getattr(info[0], '__name__', info[0]))
            if strtype in self._ignored_exceptions:
                return

            if not isinstance(info[2], bstr):
                tb_text = ''.join(
                    format_exception(*info, **{'as_html': 0}))
                tb_html = ''.join(
                    format_exception(*info, **{'as_html': 1}))
            else:
                tb_text = info[2]

            request = getattr(self, 'REQUEST', None)
            url = None
            username = None
            userid = None
            req_html = None
            try:
                strv = str(info[1])
            except Exception:
                strv = '<unprintable %s object>' % type(info[1]).__name__
            if request:
                url = request.get('URL', '?')
                usr = getSecurityManager().getUser()
                username = usr.getUserName()
                userid = usr.getId()
                try:
                    req_html = str(request)
                except Exception:
                    pass
                if strtype == 'NotFound':
                    strv = url
                    next = request['TraversalRequestNameStack']
                    if next:
                        next = list(next)
                        next.reverse()
                        strv = '%s [ /%s ]' % (strv, '/'.join(next))

            log = self._getLog()
            entry_id = str(now) + str(random())  # Low chance of collision
            log.append({
                'type': strtype,
                'value': strv,
                'time': now,
                'id': entry_id,
                'tb_text': tb_text,
                'tb_html': tb_html,
                'username': username,
                'userid': userid,
                'url': url,
                'req_html': req_html})

            cleanup_lock.acquire()
            try:
                if len(log) >= self.keep_entries:
                    del log[:-self.keep_entries]
            finally:
                cleanup_lock.release()
        except Exception:
            LOG.error('Error while logging', exc_info=sys.exc_info())
        else:
            notify(ErrorRaisedEvent(log[-1]))
            if self.copy_to_zlog:
                self._do_copy_to_zlog(now, strtype, entry_id,
                                      str(url), tb_text)
            return '%s/showEntry?id=%s' % (self.absolute_url(), entry_id)

    def _do_copy_to_zlog(self, now, strtype, entry_id, url, tb_text):
        when = _rate_restrict_pool.get(strtype, 0)
        if now > when:
            next_when = max(when,
                            now - _rate_restrict_burst * _rate_restrict_period)
            next_when += _rate_restrict_period
            _rate_restrict_pool[strtype] = next_when
            LOG.error('%s %s\n%s' % (entry_id, url, tb_text.rstrip()))

    @security.protected(use_error_logging)
    def getProperties(self):
        return {
            'keep_entries': self.keep_entries,
            'copy_to_zlog': self.copy_to_zlog,
            'ignored_exceptions': self._ignored_exceptions,
        }

    @security.protected(log_to_event_log)
    def checkEventLogPermission(self):
        if not getSecurityManager().checkPermission(log_to_event_log, self):
            raise Unauthorized('You do not have the "%s" permission.' %
                               log_to_event_log)
        return 1

    @security.protected(use_error_logging)
    def setProperties(self, keep_entries, copy_to_zlog=0,
                      ignored_exceptions=(), RESPONSE=None):
        """Sets the properties of this site error log.
        """
        copy_to_zlog = not not copy_to_zlog
        if copy_to_zlog and not self.copy_to_zlog:
            # Before turning on event logging, check the permission.
            self.checkEventLogPermission()
        self.keep_entries = int(keep_entries)
        self.copy_to_zlog = copy_to_zlog
        # filter out empty lines
        # ensure we don't save exception objects but exceptions instead
        self._ignored_exceptions = tuple(
            [_f for _f in map(str, ignored_exceptions) if _f])
        if RESPONSE is not None:
            RESPONSE.redirect(
                '%s/manage_main?manage_tabs_message=Changed+properties.' %
                self.absolute_url())

    @security.protected(use_error_logging)
    def getLogEntries(self):
        """Returns the entries in the log, most recent first.

        Makes a copy to prevent changes.
        """
        # List incomprehension ;-)
        res = [entry.copy() for entry in self._getLog()]
        res.reverse()
        return res

    @security.protected(use_error_logging)
    def getLogEntryById(self, id):
        """Returns the specified log entry.

        Makes a copy to prevent changes.  Returns None if not found.
        """
        for entry in self._getLog():
            if entry['id'] == id:
                return entry.copy()
        return None

    @security.protected(use_error_logging)
    def getLogEntryAsText(self, id, RESPONSE=None):
        """Returns the specified log entry.

        Makes a copy to prevent changes.  Returns None if not found.
        """
        entry = self.getLogEntryById(id)
        if entry is None:
            return 'Log entry not found or expired'
        if RESPONSE is not None:
            RESPONSE.setHeader('Content-Type', 'text/plain')
        return entry['tb_text']


InitializeClass(SiteErrorLog)


def manage_addErrorLog(dispatcher, RESPONSE=None):
    """Add a site error log to a container."""
    log = SiteErrorLog()
    dispatcher._setObject(log.id, log)
    if RESPONSE is not None:
        RESPONSE.redirect(
            dispatcher.DestinationURL()
            + '/manage_main?manage_tabs_message=Error+Log+Added.')


@adapter(IPubFailure)
def IPubFailureSubscriber(event):
    """ Handles an IPubFailure event triggered by the WSGI Publisher.
        This handler forwards the event to the SiteErrorLog object
        closest to the published object that the error occured with,
        it logs no error if no published object was found.
    """
    request = event.request
    published = request.get('PUBLISHED')
    if published is None:  # likely a traversal problem
        parents = request.get('PARENTS')
        if parents:
            # partially emulate successful traversal
            published = request['PUBLISHED'] = parents.pop(0)
    if published is None:
        return

    published = getattr(published, '__self__', published)  # method --> object
    try:
        error_log = aq_acquire(published, '__error_log__', containment=1)
    except AttributeError:
        pass
    else:
        error_log.raising(event.exc_info)



##############################################################################
#
# Copyright (c) 2001, 2002 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

import logging
import sys
import unittest

import Testing.testbrowser
import Testing.ZopeTestCase
import transaction
import Zope2
import Zope2.App
from OFS.Folder import Folder
from OFS.Folder import manage_addFolder
from Testing.makerequest import makerequest
from zope.component import adapter
from zope.component import provideHandler
from zope.component.globalregistry import globalSiteManager
from zope.event import notify
from ZPublisher.pubevents import PubFailure
from ZPublisher.WSGIPublisher import publish
from ZPublisher.WSGIPublisher import transaction_pubevents

from ..interfaces import IErrorRaisedEvent
from ..SiteErrorLog import IPubFailureSubscriber
from ..SiteErrorLog import manage_addErrorLog


class SiteErrorLogTests(unittest.TestCase):

    def setUp(self):
        Zope2.startup_wsgi()
        transaction.begin()
        self.app = makerequest(Zope2.app())
        try:
            if not hasattr(self.app, 'error_log'):
                # If ZopeLite was imported, we have no default error_log
                from Products.SiteErrorLog.SiteErrorLog import SiteErrorLog
                self.app._setObject('error_log', SiteErrorLog())
            self.app.manage_addDTMLMethod('doc', '')

            self.logger = logging.getLogger('Zope.SiteErrorLog')
            self.log = logging.handlers.BufferingHandler(sys.maxsize)
            self.logger.addHandler(self.log)
            self.old_level = self.logger.level
            self.logger.setLevel(logging.ERROR)
        except Exception:
            self.tearDown()

    def tearDown(self):
        self.logger.removeHandler(self.log)
        self.logger.setLevel(self.old_level)
        transaction.abort()
        self.app._p_jar.close()

    def testInstantiation(self):
        # Retrieve the error_log by ID
        sel_ob = getattr(self.app, 'error_log', None)

        # Does the error log exist?
        self.assertTrue(sel_ob is not None)

        # Is the __error_log__ hook in place?
        self.assertEqual(self.app.__error_log__, sel_ob)

    def testSimpleException(self):
        # Grab the Site Error Log and make sure it's empty
        sel_ob = self.app.error_log
        previous_log_length = len(sel_ob.getLogEntries())

        # Fill the DTML method at self.root.doc with bogus code
        dmeth = self.app.doc
        dmeth.manage_upload(file="""<dtml-var expr="1/0">""")

        # Faking the behavior of the WSGIPublisher (object acquisition,
        # view calling and failure notification on exception).
        try:
            dmeth.__call__()
        except ZeroDivisionError:
            self.app.REQUEST['PUBLISHED'] = dmeth
            notify(PubFailure(self.app.REQUEST, sys.exc_info(), False))

        # Now look at the SiteErrorLog, it has one more log entry
        self.assertEqual(len(sel_ob.getLogEntries()), previous_log_length + 1)

    def testEventSubscription(self):
        # Fill the DTML method at self.root.doc with bogus code
        dmeth = self.app.doc
        dmeth.manage_upload(file="""<dtml-var expr="1/0">""")

        event_logs = []

        @adapter(IErrorRaisedEvent)
        def notifyError(evt):
            event_logs.append(evt)

        provideHandler(notifyError)
        # Faking the behavior of the WSGIPublisher (object acquisition,
        # view calling and failure notification on exception).
        try:
            dmeth.__call__()
        except ZeroDivisionError:
            self.app.REQUEST['PUBLISHED'] = dmeth
            notify(PubFailure(self.app.REQUEST, sys.exc_info(), False))

        self.assertEqual(len(event_logs), 1)
        self.assertEqual(event_logs[0]['type'], 'ZeroDivisionError')
        self.assertEqual(event_logs[0]['username'], 'Anonymous User')

    def testForgetException(self):
        elog = self.app.error_log

        # Create a predictable error
        try:
            raise AttributeError('DummyAttribute')
        except AttributeError:
            self.app.REQUEST['PUBLISHED'] = elog
            notify(PubFailure(self.app.REQUEST, sys.exc_info(), False))

        previous_log_length = len(elog.getLogEntries())

        entries = elog.getLogEntries()
        self.assertEqual(entries[0]['value'], 'DummyAttribute')

        # Kick it
        elog.forgetEntry(entries[0]['id'])

        # Really gone?
        self.assertEqual(len(elog.getLogEntries()), previous_log_length - 1)

    def testIgnoredException(self):
        # Grab the Site Error Log
        sel_ob = self.app.error_log
        previous_log_length = len(sel_ob.getLogEntries())

        # Tell the SiteErrorLog to ignore ZeroDivisionErrors
        current_props = sel_ob.getProperties()
        ignored = list(current_props['ignored_exceptions'])
        ignored.append('ZeroDivisionError')
        sel_ob.setProperties(current_props['keep_entries'],
                             copy_to_zlog=current_props['copy_to_zlog'],
                             ignored_exceptions=ignored)

        # Fill the DTML method at self.root.doc with bogus code
        dmeth = self.app.doc
        dmeth.manage_upload(file="""<dtml-var expr="1/0">""")

        # Faking the behavior of the WSGIPublisher (object acquisition,
        # view calling and failure notification on exception).
        try:
            dmeth.__call__()
        except ZeroDivisionError:
            self.app.REQUEST['PUBLISHED'] = dmeth
            notify(PubFailure(self.app.REQUEST, sys.exc_info(), False))

        # Now look at the SiteErrorLog, it must have the same number of
        # log entries
        self.assertEqual(len(sel_ob.getLogEntries()), previous_log_length)

    def testEntryID(self):
        elog = self.app.error_log

        # Create a predictable error
        try:
            raise AttributeError('DummyAttribute')
        except AttributeError:
            self.app.REQUEST['PUBLISHED'] = elog
            notify(PubFailure(self.app.REQUEST, sys.exc_info(), False))

        entries = elog.getLogEntries()
        entry_id = entries[0]['id']

        self.assertTrue(entry_id in self.log.buffer[-1].msg,
                        (entry_id, self.log.buffer[-1].msg))

    def testCleanup(self):
        # Need to make sure that the __error_log__ hook gets cleaned up
        self.app._delObject('error_log')
        self.assertEqual(getattr(self.app, '__error_log__', None), None)


class SiteErrorLogUITests(Testing.ZopeTestCase.FunctionalTestCase):

    def setUp(self):
        super(SiteErrorLogUITests, self).setUp()

        Zope2.App.zcml.load_site(force=True)

        uf = self.app.acl_users
        uf.userFolderAddUser('manager', 'manager_pass', ['Manager'], [])

        # Why is this neccessary, shouldn't the test get a new app every time?
        if not hasattr(self.app, 'error_log'):
            manage_addErrorLog(self.app)
        self.error_log = self.app.error_log

        self.browser = Testing.testbrowser.Browser()
        self.browser.login('manager', 'manager_pass')
        self.browser.open('http://localhost/error_log/manage_main')

    def testSubmitRetainsIgnoredExceptionsUnchanged(self):
        # Checks the fix for
        # https://github.com/zopefoundation/Products.SiteErrorLog/issues/13
        ignoredExceptions = self.browser.getControl(
            label='Ignored exception types')
        self.assertEqual(
            ignoredExceptions.value,
            'Unauthorized\nNotFound\nRedirect')  # default value
        ignoredExceptions.value = 'Unauthorized\nFnord'
        self.assertNotIn('Changed properties', self.browser.contents)
        self.browser.getControl('Save Changes').click()
        self.assertIn('Changed properties', self.browser.contents)
        ignoredExceptions = self.browser.getControl(
            label='Ignored exception types')
        self.assertEqual(ignoredExceptions.value, 'Unauthorized\nFnord')


class WsgiErrorlogIntegrationLayer(Testing.ZopeTestCase.layer.ZopeLite):
    """The tests using this layer commit transactions. Therefore,
    we avoid persistent changes there and build the complete
    support structure in this layer.
    """
    @classmethod
    def setUp(cls):
        # Apparently, other tests have already registered the
        #   handler below, even twice
        #   Clean up those registrations and
        #   make a single clean registration
        #   remember when we must remove our registration
        regs = [r for r in globalSiteManager.registeredHandlers()
                if r.factory is IPubFailureSubscriber
                ]
        if regs:
            globalSiteManager.unregisterHandler(IPubFailureSubscriber)
        cls._unregister = not regs
        globalSiteManager.registerHandler(IPubFailureSubscriber)
        # Set up our test structure
        #   /
        #     sel_f1/
        #             error_log
        #             sel_f2/
        #                    error_log
        app = Testing.ZopeTestCase.app()
        # first level folder
        manage_addFolder(app, 'sel_f1')
        sel_f1 = app.sel_f1
        # second level folder
        manage_addFolder(sel_f1, 'sel_f2')
        sel_f2 = sel_f1.sel_f2
        # put an error log in each of those folders
        # (used in `test_correct_log_*`)
        for f in (sel_f1, sel_f2):
            manage_addErrorLog(f)
            el = f.error_log
            el._ignored_exceptions = ()  # do not ignore exceptions
        transaction.commit()
        # make `manage_delObjects` temporarily public
        cls._saved_roles = Folder.manage_delObjects__roles__
        Folder.manage_delObjects__roles__ = None  # public

    @classmethod
    def tearDown(cls):
        Folder.manage_delObjects__roles__ = cls._saved_roles
        if cls._unregister:
            globalSiteManager.unregisterHandler(IPubFailureSubscriber)
        app = Testing.ZopeTestCase.app()
        app._delOb('sel_f1')
        transaction.commit()


class WsgiErrorlogIntegrationTests(Testing.ZopeTestCase.ZopeTestCase):
    layer = WsgiErrorlogIntegrationLayer

    # we override `ZopeTestCase.setUp` by purpose
    def setUp(self):
        app = self.app = Testing.ZopeTestCase.app()
        self.f1 = app.sel_f1
        self.f2 = self.f1.sel_f2
        self.el1 = self.f1.error_log
        self.el2 = self.f2.error_log

    # overridden by purpose
    def tearDown(self):
        self._clear_els()

    def _clear_els(self):
        for el in (self.el1, self.el2):
            el._getLog()[:] = []

    def _get_el_nos(self):
        return tuple(len(el._getLog()) for el in (self.el1, self.el2))

    def test_correct_log_traversal_2(self):
        self._request('sel_f1/sel_f2/missing')
        self.assertEqual(self._get_el_nos(), (0, 1))

    def test_correct_log_traversal_1(self):
        self._request('sel_f1/missing')
        self.assertEqual(self._get_el_nos(), (1, 0))

    def test_correct_log_method_2(self):
        self._request('sel_f1/sel_f2/manage_delObjects', dict(id='missing'))
        self.assertEqual(self._get_el_nos(), (0, 1))

    def test_correct_log_method_1(self):
        self._request('sel_f1/manage_delObjects', dict(id='missing'))
        self.assertEqual(self._get_el_nos(), (1, 0))

    def _request(self, url, params=None):
        app = self.app
        request = app.REQUEST
        request.environ['PATH_INFO'] = url
        if params:
            request.form.update(params)
            request.other.update(params)
        try:
            with transaction_pubevents(request, request.response):
                publish(request, (app, 'Zope', False))
        except Exception:
            pass
